@@ -2,7 +2,7 @@
  * ccache -- a fast C/C++ compiler cache
  *
  * Copyright (C) 2002-2007 Andrew Tridgell
- * Copyright (C) 2009-2011 Joel Rosdahl
+ * Copyright (C) 2009-2012 Joel Rosdahl
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -267,8 +267,7 @@ get_path_in_cache(const char *name, const char *suffix)
 		free(path);
 		path = p;
 		if (create_dir(path) != 0) {
-			cc_log("Failed to create %s: %s", path, strerror(errno));
-			failed();
+			fatal("Failed to create %s: %s", path, strerror(errno));
 		}
 	}
 	result = format("%s/%s%s", path, name + nlevels, suffix);
@@ -530,7 +529,10 @@ to_cache(struct args *args)
 	status = execute(args->argv, tmp_stdout, tmp_stderr);
 	args_pop(args, 3);
 
-	if (stat(tmp_stdout, &st) != 0 || st.st_size != 0) {
+	if (stat(tmp_stdout, &st) != 0) {
+		fatal("Could not create %s (permission denied?)", tmp_stdout);
+	}
+	if (st.st_size != 0) {
 		cc_log("Compiler produced stdout");
 		stats_update(STATS_STDOUT);
 		tmp_unlink(tmp_stdout);
@@ -790,6 +792,36 @@ update_cached_result_globals(struct file_hash *hash)
 	free(object_name);
 }
 
+/*
+ * Hash mtime or content of a file, or the output of a command, according to
+ * the CCACHE_COMPILERCHECK setting.
+ */
+static void
+hash_compiler(struct mdfour *hash, struct stat *st, const char *path,
+              bool allow_command)
+{
+	const char *compilercheck;
+
+	compilercheck = getenv("CCACHE_COMPILERCHECK");
+	if (!compilercheck) {
+		compilercheck = "mtime";
+	}
+	if (str_eq(compilercheck, "none")) {
+		/* Do nothing. */
+	} else if (str_eq(compilercheck, "mtime")) {
+		hash_delimiter(hash, "cc_mtime");
+		hash_int(hash, st->st_size);
+		hash_int(hash, st->st_mtime);
+	} else if (str_eq(compilercheck, "content") || !allow_command) {
+		hash_delimiter(hash, "cc_content");
+		hash_file(hash, path);
+	} else { /* command string */
+		if (!hash_multicommand_output(hash, compilercheck, orig_args->argv[0])) {
+			fatal("Failure running compiler check command: %s", compilercheck);
+		}
+	}
+}
+
 /*
  * Update a hash sum with information common for the direct and preprocessor
  * modes.
@@ -798,7 +830,6 @@ static void
 calculate_common_hash(struct args *args, struct mdfour *hash)
 {
 	struct stat st;
-	const char *compilercheck;
 	char *p;
 
 	hash_string(hash, HASH_PREFIX);
@@ -819,31 +850,16 @@ calculate_common_hash(struct args *args, struct mdfour *hash)
 	/*
 	 * Hash information about the compiler.
 	 */
-	compilercheck = getenv("CCACHE_COMPILERCHECK");
-	if (!compilercheck) {
-		compilercheck = "mtime";
-	}
-	if (str_eq(compilercheck, "none")) {
-		/* Do nothing. */
-	} else if (str_eq(compilercheck, "content")) {
-		hash_delimiter(hash, "cc_content");
-		hash_file(hash, args->argv[0]);
-	} else if (str_eq(compilercheck, "mtime")) {
-		hash_delimiter(hash, "cc_mtime");
-		hash_int(hash, st.st_size);
-		hash_int(hash, st.st_mtime);
-	} else { /* command string */
-		if (!hash_multicommand_output(hash, compilercheck, orig_args->argv[0])) {
-			fatal("Failure running compiler check command: %s", compilercheck);
-		}
-	}
+	hash_compiler(hash, &st, args->argv[0], true);
 
 	/*
 	 * Also hash the compiler name as some compilers use hard links and
 	 * behave differently depending on the real name.
 	 */
 	hash_delimiter(hash, "cc_name");
-	hash_string(hash, basename(args->argv[0]));
+	p = basename(args->argv[0]);
+	hash_string(hash, p);
+	free(p);
 
 	/* Possibly hash the current working directory. */
 	if (getenv("CCACHE_HASHDIR")) {
@@ -886,6 +902,7 @@ calculate_object_hash(struct args *args, struct mdfour *hash, int direct_mode)
 	struct stat st;
 	int result;
 	struct file_hash *object_hash = NULL;
+	char *p;
 
 	/* first the arguments */
 	for (i = 1; i < args->argc; i++) {
@@ -913,14 +930,24 @@ calculate_object_hash(struct args *args, struct mdfour *hash, int direct_mode)
 			}
 		}
 
-		if (str_startswith(args->argv[i], "--specs=") &&
-		    stat(args->argv[i] + 8, &st) == 0) {
-			/* If given a explicit specs file, then hash that file,
+		p = NULL;
+		if (str_startswith(args->argv[i], "-specs=")) {
+			p = args->argv[i] + 7;
+		} else if (str_startswith(args->argv[i], "--specs=")) {
+			p = args->argv[i] + 8;
+		}
+		if (p && stat(p, &st) == 0) {
+			/* If given an explicit specs file, then hash that file,
 			   but don't include the path to it in the hash. */
 			hash_delimiter(hash, "specs");
-			if (!hash_file(hash, args->argv[i] + 8)) {
-				failed();
-			}
+			hash_compiler(hash, &st, p, false);
+			continue;
+		}
+
+		if (str_startswith(args->argv[i], "-fplugin=")
+		    && stat(args->argv[i] + 9, &st) == 0) {
+			hash_delimiter(hash, "plugin");
+			hash_compiler(hash, &st, args->argv[i] + 9, false);
 			continue;
 		}
 
@@ -930,6 +957,24 @@ calculate_object_hash(struct args *args, struct mdfour *hash, int direct_mode)
 	}
 
 	if (direct_mode) {
+		/* Hash environment variables that affect the preprocessor output. */
+		const char **p;
+		const char *envvars[] = {
+			"CPATH",
+			"C_INCLUDE_PATH",
+			"CPLUS_INCLUDE_PATH",
+			"OBJC_INCLUDE_PATH",
+			"OBJCPLUS_INCLUDE_PATH", /* clang */
+			NULL
+		};
+		for (p = envvars; *p != NULL ; ++p) {
+			char *v = getenv(*p);
+			if (v) {
+				hash_delimiter(hash, *p);
+				hash_string(hash, v);
+			}
+		}
+
 		if (!(sloppiness & SLOPPY_FILE_MACRO)) {
 			/*
 			 * The source code file or an include file may contain
@@ -1389,6 +1434,7 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 			continue;
 		}
 		if (str_startswith(argv[i], "-MQ") || str_startswith(argv[i], "-MT")) {
+			dependency_target_specified = true;
 			args_add(dep_args, argv[i]);
 			if (strlen(argv[i]) == 3) {
 				/* -MQ arg or -MT arg */
@@ -1400,13 +1446,6 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 				}
 				args_add(dep_args, argv[i + 1]);
 				i++;
-				/*
-				 * Yes, that's right. It's strange, but apparently, GCC behaves
-				 * differently for -MT arg and -MTarg (and similar for -MQ): in the
-				 * latter case, but not in the former, an implicit dependency for the
-				 * object file is added to the dependency file.
-				 */
-				dependency_target_specified = true;
 			}
 			continue;
 		}
@@ -1824,6 +1863,14 @@ ccache(int argc, char *argv[])
 		failed();
 	}
 
+	if (!getenv("CCACHE_READONLY")) {
+		if (create_cachedirtag(cache_dir) != 0) {
+			cc_log("failed to create %s/CACHEDIR.TAG (%s)\n",
+			        cache_dir, strerror(errno));
+			failed();
+		}
+	}
+
 	sloppiness = parse_sloppiness(getenv("CCACHE_SLOPPINESS"));
 
 	cc_log_argv("Command line: ", argv);
@@ -1835,7 +1882,7 @@ ccache(int argc, char *argv[])
 	}
 
 	if (getenv("CCACHE_UNIFY")) {
-		cc_log("Unify mode disabled");
+		cc_log("Unify mode enabled");
 		enable_unify = true;
 	}
 
@@ -2176,15 +2223,6 @@ ccache_main(int argc, char *argv[])
 		exit(1);
 	}
 
-	if (!getenv("CCACHE_READONLY")) {
-		if (create_cachedirtag(cache_dir) != 0) {
-			fprintf(stderr,
-			        "ccache: failed to create %s/CACHEDIR.TAG (%s)\n",
-			        cache_dir, strerror(errno));
-			exit(1);
-		}
-	}
-
 	ccache(argc, argv);
 	return 1;
 }
